@@ -20,7 +20,7 @@ use std::{
 
 use btoi::btoi;
 use libc::{c_int, c_ulong, prctl, EACCES, EOPNOTSUPP, RTLD_LOCAL, RTLD_NOLOAD, RTLD_NOW};
-use libloading::os::unix::Library;
+use libloading::{os::unix::Library, Error as LibraryError};
 use libseccomp::{scmp_cmp, ScmpAction, ScmpArch, ScmpFilterContext, ScmpSyscall};
 use nix::{
     dir::Dir,
@@ -40,18 +40,20 @@ use serde::Serialize;
 
 use crate::{
     caps,
-    compat::{fstatx, waitid, WaitStatus, STATX_UID},
-    config::{FADVISE_SYSCALLS, HAVE_RWF_NOAPPEND, MMAP_MIN_ADDR, UNSAFE_PERSONAS},
+    compat::{waitid, WaitStatus},
+    config::{FADVISE_SYSCALLS, HAVE_RWF_NOAPPEND, MMAP_MIN_ADDR, SYD_MADVISE, UNSAFE_PERSONAS},
     cookie::safe_socket,
     err::{err2no, SydResult},
-    fs::{nlmsg_align, readlinkat, retry_on_eintr, safe_clone, FileType, AT_BADFD},
+    fs::{nlmsg_align, readlinkat, safe_clone, AT_BADFD},
     info,
     landlock::{
         path_beneath_rules, Access, AccessFs, AccessNet, NetPort, RestrictSelfFlags,
         RestrictionStatus, Ruleset, RulesetAttr, RulesetCreatedAttr, RulesetError, RulesetStatus,
         Scope, ABI,
     },
+    lookup::FileType,
     path::{mask_path, XPathBuf},
+    retry::retry_on_eintr,
     sandbox::{IoctlMap, Sandbox},
 };
 
@@ -444,7 +446,7 @@ pub fn confine_scmp_write(
         }
     };
 
-    if cfg!(feature = "prof") && (chk_mem && Sandbox::memory_access() < 2) {
+    if cfg!(feature = "prof") || (chk_mem && Sandbox::memory_access() < 2) {
         ctx.add_rule(ScmpAction::Allow, syscall)?;
         return Ok(());
     }
@@ -479,6 +481,20 @@ pub fn confine_scmp_fadvise(ctx: &mut ScmpFilterContext) -> SydResult<()> {
     Ok(())
 }
 
+/// Confine madvise(2) advice.
+pub fn confine_scmp_madvise(ctx: &mut ScmpFilterContext) -> SydResult<()> {
+    if let Ok(syscall) = ScmpSyscall::from_name("madvise") {
+        for advice in SYD_MADVISE {
+            ctx.add_rule_conditional(ScmpAction::Allow, syscall, &[scmp_cmp!($arg2 == *advice)])?;
+        }
+    } else {
+        info!("ctx": "confine", "op": "allow_syscall",
+            "msg": "invalid or unsupported syscall madvise");
+    }
+
+    Ok(())
+}
+
 /// Convenience `Command` run wrapper which returns:
 ///
 /// - Same exit code as the process on clean exit.
@@ -867,44 +883,39 @@ pub(crate) fn op2name(op: u8) -> &'static str {
         0x5 => "accept",
         0x6 => "getsockname",
         0x7 => "getpeername",
+        0x8 => "socketpair",
         0x9 => "send",
         0xb => "sendto",
+        0xc => "recvfrom",
         0xf => "getsockopt",
         0x10 => "sendmsg",
+        0x11 => "recvmsg",
         0x12 => "accept4",
+        0x13 => "recvmmsg",
+        u8::MAX => "recvmmsg_time64",
         0x14 => "sendmmsg",
         _ => unreachable!("BUG: op2name called with unsupported op:{op:#x}, report a bug!"),
     }
 }
 
-#[inline]
 pub(crate) fn op2errno(op: u8) -> Errno {
+    // We used to return:
+    // 0x2 = bind = EADDRNOTAVAIL.
+    // 0x3 = connect = ECONNREFUSED.
+    // 0xb,0x10,0x14 = send{to,{m,}msg} = ENOTCONN.
+    // As of version 3.41.8, this was changed to EACCES.
+    // This is consistent with SELinux.
+    // EACCES is not a valid errno(3) for accept{,4}.
+    // EACCES is not a valid errno(3) for socketpair(2).
     match op {
-        0x2 /*bind*/ => Errno::EADDRNOTAVAIL,
-        0x3 /*connect*/ =>  Errno::ECONNREFUSED,
         0x5 | 0x12 /*accept{,4}*/ => Errno::ECONNABORTED,
-        _ /*send{to,{m,}msg}*/ => Errno::ENOTCONN,
+        0x8 /* socketpair */ => Errno::EOPNOTSUPP,
+        _ => Errno::EACCES,
     }
 }
 
 /// Checks if the given namespaces are enabled.
 pub fn ns_enabled(ns_flags: CloneFlags) -> Result<bool, Errno> {
-    // Quick path: check for docker:
-    // Is /proc owned by root?
-    #[expect(clippy::disallowed_methods)]
-    if nix::fcntl::openat(
-        AT_BADFD,
-        "/proc",
-        OFlag::O_RDONLY | OFlag::O_DIRECTORY,
-        Mode::empty(),
-    )
-    .and_then(|fd| fstatx(fd, STATX_UID))
-    .map(|statx| statx.stx_uid != 0)
-    .unwrap_or(false)
-    {
-        return Ok(false);
-    }
-
     // CLONE_SIGHAND|CLONE_V{FORK,M} are not included intentionally.
     const SAFE_CLONE_FLAGS: libc::c_int = libc::CLONE_FS | libc::CLONE_FILES | libc::CLONE_IO;
 
@@ -1798,12 +1809,7 @@ pub fn vdso_list_calls() -> Result<Vec<&'static CStr>, libloading::Error> {
         c"__vdso_time",
     ];
 
-    // SAFETY: Library::open with RTLD_NOLOAD only obtains a handle
-    // to the already-mapped vDSO; it does not load code. Handle used
-    // immediately and dropped.
-    let vdso =
-        unsafe { Library::open(Some("linux-vdso.so.1"), RTLD_NOLOAD | RTLD_LOCAL | RTLD_NOW) }?;
-
+    let vdso = vdso_open()?;
     let mut out = Vec::with_capacity(VDSO_CALL_NAMES.len());
     for (idx, sym) in VDSO_CALL_NAMES.iter().enumerate() {
         let sym = sym.to_bytes_with_nul();
@@ -1826,12 +1832,7 @@ pub fn vdso_list_calls() -> Result<Vec<&'static CStr>, libloading::Error> {
 ///
 /// Returns true if the vDSO object exports the symbol.
 pub fn has_vdso_symbol(sym: &CStr) -> bool {
-    // SAFETY: Library::open with RTLD_NOLOAD only obtains a handle
-    // to the already-mapped vDSO; it does not load code. Handle used
-    // immediately and dropped.
-    let vdso = if let Ok(vdso) =
-        unsafe { Library::open(Some("linux-vdso.so.1"), RTLD_NOLOAD | RTLD_LOCAL | RTLD_NOW) }
-    {
+    let vdso = if let Ok(vdso) = vdso_open() {
         vdso
     } else {
         return false;
@@ -1842,6 +1843,24 @@ pub fn has_vdso_symbol(sym: &CStr) -> bool {
     unsafe { vdso.get::<*const ()>(sym.to_bytes_with_nul()).is_ok() }
 }
 
+/// Open vDSO and return the Library instance.
+///
+/// Attempts to load `linux-vdso.so.1` first, and falls back to `linux-vdso64.so.1` (ppc64).
+pub fn vdso_open() -> Result<Library, LibraryError> {
+    match vdso_open_by_name("linux-vdso.so.1") {
+        Ok(lib) => Ok(lib),
+        Err(LibraryError::DlOpen { .. }) => vdso_open_by_name("linux-vdso64.so.1"),
+        Err(error) => Err(error),
+    }
+}
+
+/// Open vDSO by name and return the Library instance.
+pub fn vdso_open_by_name(name: &str) -> Result<Library, LibraryError> {
+    // SAFETY: Library::open with RTLD_NOLOAD only obtains a handle
+    // to the already-mapped vDSO; it does not load code.
+    unsafe { Library::open(Some(name), RTLD_NOLOAD | RTLD_LOCAL | RTLD_NOW) }
+}
+
 /// Check support for CONFIG_UNIX_DIAG in Linux kernel.
 #[expect(clippy::arithmetic_side_effects)]
 #[expect(clippy::cast_possible_truncation)]
