@@ -34,7 +34,7 @@ pub(crate) fn sys_ioctl(request: UNotifyEventRequest) -> ScmpNotifResp {
             fsflags: FsFlags::MUST_PATH | FsFlags::WANT_READ,
             ..Default::default()
         };
-        let (path, _, _) = request.read_path(&sandbox, arg, false)?;
+        let (path, _, _) = request.read_path(&sandbox, arg)?;
 
         // Check if the ioctl(2) request is allowlisted or denylisted.
         let arg = req.data.args[1];
