@@ -10,7 +10,6 @@ macro_rules! syscall_handler {
     ($request:ident, $body:expr) => {{
         let request_id = $request.scmpreq.id;
 
-        #[expect(clippy::arithmetic_side_effects)]
         match $body($request) {
             Ok(result) => result,
             // SAFETY: Harden against UnknownErrno so as not to
@@ -19,7 +18,11 @@ macro_rules! syscall_handler {
             // SAFETY: ECANCELED is used by Syd internally to denote
             // requests that should be turned into no-op.
             Err(Errno::ECANCELED) => ScmpNotifResp::new(request_id, 0, 0, 0),
-            Err(errno) => ScmpNotifResp::new(request_id, 0, -(errno as i32), 0),
+            Err(errno) => {
+                // `ScmpNotifResp` expects negated errno.
+                let errno = (errno as i32).checked_neg().unwrap_or(-libc::ENOSYS);
+                ScmpNotifResp::new(request_id, 0, errno, 0)
+            }
         }
     }};
 }
@@ -81,6 +84,9 @@ pub(crate) mod open;
 /// prctl(2) handler
 pub(crate) mod prctl;
 
+/// readlink(2) and readlinkat(2) handlers
+pub(crate) mod readlink;
+
 /// rename(2), renameat(2) and renameat2(2) handlers
 pub(crate) mod rename;
 
@@ -126,10 +132,12 @@ pub(crate) mod unlink;
 /// xattr handlers
 pub(crate) mod xattr;
 
+/// ptrace(2) hooks
+pub(crate) mod ptrace;
+
 use std::borrow::Cow;
 
 use libseccomp::ScmpNotifResp;
-use memchr::memmem;
 use nix::{
     errno::Errno,
     fcntl::{AtFlags, OFlag},
@@ -141,11 +149,13 @@ use nix::{
 };
 
 use crate::{
-    fs::{fd_status_flags, to_valid_fd, CanonicalPath, FileInfo, FileType},
-    hook::{PathArgs, SysArg, UNotifyEventRequest},
+    fs::{fd_status_flags, to_valid_fd},
     ioctl::{ioctl_names_get, Ioctl},
+    log::log_is_main,
+    lookup::{CanonicalPath, FileInfo, FileType},
     notice,
     path::{XPath, XPathBuf},
+    req::{PathArgs, SysArg, UNotifyEventRequest},
     sandbox::{Action, Capability, Sandbox, SandboxGuard},
     warn,
 };
@@ -328,6 +338,8 @@ pub(crate) fn sandbox_path(
             }
         }
         Action::Deny | Action::Filter => Err(deny_errno),
+        //Do NOT panic the main thread!
+        Action::Panic if log_is_main(std::thread::current().id()) => Err(deny_errno),
         Action::Panic => panic!(),
         Action::Exit => std::process::exit(deny_errno as i32),
         Action::Stop => {
@@ -401,19 +413,14 @@ where
         let crypt = sandbox.enabled(Capability::CAP_CRYPT);
         let hide = !is_fd && sandbox.enabled(Capability::CAP_STAT);
 
-        // EXCEPTION: We do want to return success t o _access_(2) calls
-        // to magic paths in case the sandbox lock allows it.
-        let mut magic = !is_fd
-            && !sandbox.locked_for(req.pid())
-            && memmem::find(syscall_name.as_bytes(), b"access").is_some();
-
+        let mut magic = false;
         let mut paths: [Option<CanonicalPath>; 2] = [None, None];
         for (idx, arg) in path_argv.iter().enumerate() {
             // Handle system calls that take a FD only,
             // such as fchmod, fchown, falllocate, ftruncate,
             // fgetxattr, fsetxattr safely and efficiently.
             if arg.path.is_some() {
-                let (path, is_magic) = request.read_path(&sandbox, *arg, magic)?;
+                let (path, is_magic, _) = request.read_path(&sandbox, *arg)?;
                 magic = is_magic;
 
                 if sandbox.is_chroot() {
@@ -480,6 +487,11 @@ where
             }
         }
 
+        if magic && sandbox.locked_for(req.pid()) {
+            // Sandbox is locked, access denied.
+            return Err(Errno::ENOENT);
+        }
+
         if !magic {
             // Call sandbox access checker, skip magic paths.
             match (&paths[0], &paths[1]) {
@@ -593,12 +605,25 @@ pub(crate) fn to_atflags(arg: u64, valid: AtFlags) -> Result<AtFlags, Errno> {
     Ok(flags)
 }
 
+// to_mode that strips unknown bits.
 #[inline]
-pub(crate) fn to_mode(arg: u64) -> Result<Mode, Errno> {
-    let mode = arg
-        .try_into()
-        .map(|mode: libc::mode_t| mode & !SFlag::S_IFMT.bits())
-        .or(Err(Errno::EINVAL))?;
+pub(crate) fn to_mode(arg: u64) -> Mode {
+    // Linux VFS only honors these chmod bits (07777).
+    const S_IALLUGO: libc::mode_t = libc::S_ISUID
+        | libc::S_ISGID
+        | libc::S_ISVTX
+        | libc::S_IRWXU
+        | libc::S_IRWXG
+        | libc::S_IRWXO;
+
+    #[expect(clippy::cast_possible_truncation)]
+    Mode::from_bits_truncate((arg as libc::mode_t) & S_IALLUGO)
+}
+
+// to_mode that rejects unknown/invalid bits.
+#[inline]
+pub(crate) fn to_mode2(arg: u64) -> Result<Mode, Errno> {
+    let mode = arg.try_into().or(Err(Errno::EINVAL))?;
     Mode::from_bits(mode).ok_or(Errno::EINVAL)
 }
 
