@@ -9,12 +9,13 @@
 use std::os::fd::{AsFd, AsRawFd};
 
 use libseccomp::ScmpNotifResp;
+use memchr::arch::all::{is_prefix, is_suffix};
 use nix::{errno::Errno, NixPath};
 
 use crate::{
     config::MMAP_MIN_ADDR,
-    fs::{FileType, FsFlags},
     kernel::sandbox_path,
+    lookup::{FileType, FsFlags},
     magic::ProcMagic,
     path::{XPathBuf, PATH_MAX},
     proc::proc_tgid,
@@ -81,6 +82,7 @@ pub(crate) fn sys_readlinkat(request: UNotifyEventRequest) -> ScmpNotifResp {
     syscall_readlink_handler(request, arg, 2, size)
 }
 
+#[expect(clippy::cognitive_complexity)]
 fn syscall_readlink_handler(
     request: UNotifyEventRequest,
     arg: SysArg,
@@ -92,7 +94,7 @@ fn syscall_readlink_handler(
         let sandbox = request.get_sandbox();
 
         // Read the remote path.
-        let (path, _, empty_path) = request.read_path(&sandbox, arg, false)?;
+        let (path, _, empty_path) = request.read_path(&sandbox, arg)?;
 
         // Check for access, allow access to fd-only calls.
         if !empty_path && sandbox.enabled(Capability::CAP_STAT) {
@@ -150,13 +152,10 @@ fn syscall_readlink_handler(
 
             if let Some((pid, maybe_tgid)) = maybe_magic_self {
                 let buf = if let Some(tgid) = maybe_tgid {
-                    let mut buf = XPathBuf::from_pid(tgid);
-                    buf.push(b"task");
-                    buf.push_pid(pid);
-                    buf
+                    XPathBuf::from_task(tgid, pid)
                 } else {
                     XPathBuf::from_pid(pid)
-                };
+                }?;
 
                 let buf = buf.as_bytes();
                 let siz = buf.len().min(buf_siz);
@@ -195,8 +194,17 @@ fn syscall_readlink_handler(
         })
         .map(|size| size as usize)?;
 
+        // Rearrange !memfd:syd/ links.
+        let mut buf = &buf[..size];
+        if is_prefix(buf, b"/memfd:syd") {
+            buf = &buf[b"/memfd:syd".len()..];
+            if is_suffix(buf, b" (deleted)") {
+                buf = &buf[..buf.len().saturating_sub(b" (deleted)".len())];
+            }
+        }
+
         // readlink(2) truncates and does NOT add a NUL-byte.
-        let size = request.write_mem(&buf[..size], req.data.args[buf_idx])?;
+        let size = request.write_mem(buf, req.data.args[buf_idx])?;
 
         // readlink(2) system call has been successfully emulated.
         #[expect(clippy::cast_possible_wrap)]
