@@ -142,12 +142,16 @@ pub mod fs;
 pub mod hash;
 /// Secure computing hooks
 pub mod hook;
+/// I/O utilities
+pub mod io;
 /// ioctl(2) request decoder
 pub mod ioctl;
 /// Landlock policy helper library for Syd
 pub mod landlock_policy;
 /// Simple logging on standard error using JSON lines
 pub mod log;
+/// Path lookup and canonicalization utilities
+pub mod lookup;
 /// magic symlink utilities
 pub mod magic;
 /// Interface to Open File Description locks
@@ -190,6 +194,8 @@ pub mod uts;
 pub mod wildmatch;
 /// Interface to wordexp(3)
 pub mod wordexp;
+/// Extended attribute utilities
+pub mod xattr;
 
 // Vendored crates:
 /// Interface to Linux capabilities
@@ -456,7 +462,7 @@ pub fn syd_info(verbose: bool) -> Result<(), Errno> {
                 .join(", ");
             printfln!("List of vDSO calls: {names}.")?;
         }
-        Err(error) => printfln!("List of vDSO calls: ? (error: {error}")?,
+        Err(error) => printfln!("List of vDSO calls: ? (error: {error}).")?,
     }
 
     // Print information on open file limits.
@@ -481,6 +487,7 @@ pub fn syd_info(verbose: bool) -> Result<(), Errno> {
     // Print detected host Linux kernel version and related features.
     printf!("Host Linux: {}.{} with", KERNEL_VERSION.0, KERNEL_VERSION.1)?;
     printf!(" mmap_min_addr={}", *MMAP_MIN_ADDR)?;
+    printf!(", page_size={}", *PAGE_SIZE)?;
     printf!(
         ", {}at_execve_check",
         if *HAVE_AT_EXECVE_CHECK { "+" } else { "-" }
@@ -1006,8 +1013,7 @@ pub fn loopback_set_bigtcp(ifindex: i32, max_size: u32) -> Result<(), Errno> {
     #[expect(clippy::arithmetic_side_effects)]
     #[expect(clippy::cast_possible_truncation)]
     {
-        nl_hdr.nlmsg_len =
-            (std::mem::size_of::<libc::nlmsghdr>() + std::mem::size_of::<ifinfomsg>()) as u32;
+        nl_hdr.nlmsg_len = (size_of::<libc::nlmsghdr>() + size_of::<ifinfomsg>()) as u32;
         nl_hdr.nlmsg_type = libc::RTM_NEWLINK;
         nl_hdr.nlmsg_flags = (libc::NLM_F_REQUEST | libc::NLM_F_ACK) as u16;
         nl_hdr.nlmsg_seq = 1;
@@ -1039,14 +1045,13 @@ pub fn loopback_set_bigtcp(ifindex: i32, max_size: u32) -> Result<(), Errno> {
         #[expect(clippy::cast_possible_truncation)]
         unsafe {
             (*attr_ptr).nla_type = kind;
-            (*attr_ptr).nla_len =
-                (std::mem::size_of::<nlattr>() + std::mem::size_of::<u32>()) as u16;
+            (*attr_ptr).nla_len = (size_of::<nlattr>() + size_of::<u32>()) as u16;
         }
 
         // SAFETY: Write u32 payload.
         #[expect(clippy::cast_ptr_alignment)]
         unsafe {
-            let ptr = (attr_ptr as *mut u8).add(std::mem::size_of::<nlattr>()) as *mut u32;
+            let ptr = (attr_ptr as *mut u8).add(size_of::<nlattr>()) as *mut u32;
             *ptr = max_size;
         }
 
@@ -1085,26 +1090,23 @@ pub fn loopback_set_bigtcp(ifindex: i32, max_size: u32) -> Result<(), Errno> {
     let mut offset = 0;
     #[expect(clippy::arithmetic_side_effects)]
     #[expect(clippy::cast_ptr_alignment)]
-    while offset + std::mem::size_of::<libc::nlmsghdr>() <= buf.len() {
+    while offset + size_of::<libc::nlmsghdr>() <= buf.len() {
         // SAFETY: We're reading a netlink message header from validated bounds.
         let hdr = unsafe { &*(buf.as_ptr().add(offset) as *const libc::nlmsghdr) };
 
         let len = hdr.nlmsg_len as usize;
-        if len < std::mem::size_of::<libc::nlmsghdr>() || offset + len > buf.len() {
+        if len < size_of::<libc::nlmsghdr>() || offset + len > buf.len() {
             return Err(Errno::EINVAL);
         }
 
         // Check for error message.
         #[expect(clippy::cast_possible_truncation)]
         if hdr.nlmsg_type == libc::NLMSG_ERROR as libc::c_ushort
-            && len >= std::mem::size_of::<libc::nlmsghdr>() + std::mem::size_of::<libc::nlmsgerr>()
+            && len >= size_of::<libc::nlmsghdr>() + size_of::<libc::nlmsgerr>()
         {
             // SAFETY: Enough data to safely parse nlmsgerr.
             let err = unsafe {
-                &*(buf
-                    .as_ptr()
-                    .add(offset + std::mem::size_of::<libc::nlmsghdr>())
-                    as *const libc::nlmsgerr)
+                &*(buf.as_ptr().add(offset + size_of::<libc::nlmsghdr>()) as *const libc::nlmsgerr)
             };
             if err.error != 0 {
                 return Err(Errno::from_raw(-err.error));
