@@ -19,7 +19,7 @@ use std::{
         fd::{AsFd, AsRawFd, OwnedFd, RawFd},
         unix::ffi::OsStrExt,
     },
-    sync::{Arc, LazyLock, RwLock},
+    sync::{Arc, RwLock},
 };
 
 use bitflags::bitflags;
@@ -39,20 +39,18 @@ use nix::{
     NixPath,
 };
 use serde::{ser::SerializeMap, Serialize};
+use zeroize::Zeroizing;
 
 use crate::{
     compat::{fstatx, timespec_tv_nsec_t, TimeSpec32, TimeSpec64, STATX_INO, XATTR_NAME_MAX},
-    config::{
-        ENV_NO_CROSS_MEMORY_ATTACH, ENV_PROC_PID_MEM_FALLBACK, HAVE_CROSS_MEMORY_ATTACH,
-        MMAP_MIN_ADDR, PROC_FILE,
-    },
+    config::{MMAP_MIN_ADDR, PAGE_SIZE, PROC_FILE},
     confine::{op2name, scmp_arch_bits, ScmpNotifReq, SydArch, Sydcall, EIDRM},
     error,
     fs::{
-        fd_status_flags, file_type, pidfd_getfd, pidfd_open, pidfd_send_signal, process_mrelease,
-        safe_canonicalize, safe_open_msym, seccomp_notify_addfd, seccomp_notify_id_valid,
-        to_valid_fd, unix_inodes, CanonicalPath, FileType, FsFlags, PIDFD_THREAD,
+        fd_status_flags, pidfd_getfd, pidfd_open, pidfd_send_signal, process_mrelease,
+        seccomp_notify_addfd, seccomp_notify_id_valid, to_valid_fd, unix_inodes, PIDFD_THREAD,
     },
+    lookup::{file_type, safe_canonicalize, safe_open_msym, CanonicalPath, FileType, FsFlags},
     path::{XPath, XPathBuf, PATH_MAX, PATH_MIN},
     proc::{
         proc_auxv, proc_comm, proc_maps, proc_rand_fd, proc_stack_pointer, proc_status, proc_tgid,
@@ -71,6 +69,15 @@ bitflags! {
     pub(crate) struct SysFlags: u8 {
         /// Whether if it's ok for the path to be empty.
         const EMPTY_PATH = 1 << 0;
+        /// The system call should be checked for /dev/syd access.
+        const CHECK_MAGIC = 1 << 1;
+    }
+}
+
+impl SysFlags {
+    /// Return true if syscall should be checked for /dev/syd access.
+    pub fn is_check_magic(self) -> bool {
+        self.contains(Self::CHECK_MAGIC)
     }
 }
 
@@ -156,6 +163,10 @@ impl SysArg {
             fsflags |= FsFlags::RESOLVE_BENEATH;
         }
 
+        if rflags.contains(ResolveFlag::RESOLVE_IN_ROOT) {
+            fsflags |= FsFlags::RESOLVE_IN_ROOT;
+        }
+
         if rflags.contains(ResolveFlag::RESOLVE_NO_SYMLINKS) {
             fsflags |= FsFlags::NO_RESOLVE_PATH;
         }
@@ -232,7 +243,6 @@ impl RemoteProcess {
         sandbox: &SandboxGuard,
         args: [u64; 6],
         arg: SysArg,
-        mut magic: bool,
         request: Option<&UNotifyEventRequest>,
     ) -> Result<(CanonicalPath<'b>, bool, bool, bool), Errno> {
         let orig = match arg.path {
@@ -248,11 +258,9 @@ impl RemoteProcess {
         };
         let mut doterr = false;
 
-        // magic is both an in and out variable.
-        // in=t: check for magic path.
-        // out=t: path is magic path.
-        let check_magic = magic;
-        magic = false;
+        // Should we check for magic path?
+        let check_magic = arg.flags.is_check_magic();
+        let mut is_magic = false;
 
         let mut empty_path = false;
         let canonical_path = if let Some(path) = orig {
@@ -275,7 +283,7 @@ impl RemoteProcess {
             }
 
             if check_magic && path.is_magic() {
-                magic = true;
+                is_magic = true;
                 CanonicalPath::new_magic(path)
             } else if empty_path || path.is_dot() {
                 let dirfd = if let Some(idx) = arg.dirfd {
@@ -398,7 +406,7 @@ impl RemoteProcess {
             }
         };
 
-        if !magic && arg.path.is_some() {
+        if !is_magic && arg.path.is_some() {
             // SAFETY: Deny access to critical and/or suspicious paths.
             canonical_path.abs().check(
                 self.pid,
@@ -409,7 +417,109 @@ impl RemoteProcess {
             )?;
         }
 
-        Ok((canonical_path, magic, doterr, empty_path))
+        Ok((canonical_path, is_magic, doterr, empty_path))
+    }
+
+    /// Allocate and read a `Zeroizing` buffer from remote process's memory with `process_vm_readv()`.
+    ///
+    /// # Safety
+    ///
+    /// This function is unsafe because the request is not validated.
+    pub(crate) unsafe fn read_vec_zeroed(
+        &self,
+        remote_addr: u64,
+        len: usize,
+    ) -> Result<Zeroizing<Vec<u8>>, Errno> {
+        if Sandbox::use_proc_pid_mem() {
+            return self.read_vec_zeroed_proc(remote_addr, len);
+        }
+
+        let mut local_buffer = Zeroizing::new(Vec::new());
+
+        // Check for zero length and return an empty Vector.
+        if len == 0 {
+            return Ok(local_buffer);
+        }
+
+        // SAFETY: Check pointer against mmap_min_addr before allocation,
+        // but after length is zero check.
+        if remote_addr < *MMAP_MIN_ADDR {
+            return Err(Errno::EFAULT);
+        }
+        let remote_addr = usize::try_from(remote_addr).or(Err(Errno::EFAULT))?;
+
+        local_buffer.try_reserve(len).or(Err(Errno::ENOMEM))?;
+
+        // SAFETY: We are using `set_len(len)` after reserving enough
+        // capacity with `try_reserve(len)`. This is safe because the
+        // buffer was already allocated with enough memory to hold `len`
+        // elements, and we are not exceeding the reserved capacity. The
+        // memory is valid for `len` elements.
+        unsafe { local_buffer.set_len(len) };
+
+        let len = process_vm_readv(
+            self.pid,
+            &mut [IoSliceMut::new(&mut local_buffer)],
+            &[RemoteIoVec {
+                len,
+                base: remote_addr,
+            }],
+        )?;
+
+        // SAFETY: len is returned by the Linux kernel.
+        unsafe { local_buffer.set_len(len) };
+        local_buffer.shrink_to_fit();
+
+        Ok(local_buffer)
+    }
+
+    /// Allocate and read a buffer from remote process's memory with `process_vm_readv()`.
+    ///
+    /// # Safety
+    ///
+    /// This function is unsafe because the request is not validated.
+    pub(crate) unsafe fn read_vec(&self, remote_addr: u64, len: usize) -> Result<Vec<u8>, Errno> {
+        if Sandbox::use_proc_pid_mem() {
+            return self.read_vec_proc(remote_addr, len);
+        }
+
+        let mut local_buffer = Vec::new();
+
+        // Check for zero length and return an empty Vector.
+        if len == 0 {
+            return Ok(local_buffer);
+        }
+
+        // SAFETY: Check pointer against mmap_min_addr before allocation,
+        // but after length is zero check.
+        if remote_addr < *MMAP_MIN_ADDR {
+            return Err(Errno::EFAULT);
+        }
+        let remote_addr = usize::try_from(remote_addr).or(Err(Errno::EFAULT))?;
+
+        local_buffer.try_reserve(len).or(Err(Errno::ENOMEM))?;
+
+        // SAFETY: We are using `set_len(len)` after reserving enough
+        // capacity with `try_reserve(len)`. This is safe because the
+        // buffer was already allocated with enough memory to hold `len`
+        // elements, and we are not exceeding the reserved capacity. The
+        // memory is valid for `len` elements.
+        unsafe { local_buffer.set_len(len) };
+
+        let len = process_vm_readv(
+            self.pid,
+            &mut [IoSliceMut::new(&mut local_buffer)],
+            &[RemoteIoVec {
+                len,
+                base: remote_addr,
+            }],
+        )?;
+
+        // SAFETY: len is returned by the Linux kernel.
+        unsafe { local_buffer.set_len(len) };
+        local_buffer.shrink_to_fit();
+
+        Ok(local_buffer)
     }
 
     /// Read data from remote process's memory with `process_vm_readv()`.
@@ -421,33 +531,148 @@ impl RemoteProcess {
         &self,
         local_buffer: &mut [u8],
         remote_addr: u64,
+        len: usize,
     ) -> Result<usize, Errno> {
-        static FORCE_PROC: LazyLock<bool> =
-            LazyLock::new(|| std::env::var_os(ENV_NO_CROSS_MEMORY_ATTACH).is_some());
-        // SAFETY: Falling back to proc_pid_mem(5) is insecure,
-        // and is no longer the default as of 3.32.6.
-        static PROC_MEM_F: LazyLock<bool> =
-            LazyLock::new(|| std::env::var_os(ENV_PROC_PID_MEM_FALLBACK).is_some());
-        if *FORCE_PROC || (*PROC_MEM_F && !*HAVE_CROSS_MEMORY_ATTACH) {
-            return self.read_mem_proc(local_buffer, remote_addr);
+        if Sandbox::use_proc_pid_mem() {
+            return self.read_mem_proc(local_buffer, remote_addr, len);
         }
 
         // SAFETY: Check pointer against mmap_min_addr.
         if remote_addr < *MMAP_MIN_ADDR {
             return Err(Errno::EFAULT);
         }
+        let remote_addr = usize::try_from(remote_addr).or(Err(Errno::EFAULT))?;
 
-        let len = local_buffer.len();
         process_vm_readv(
             self.pid,
             &mut [IoSliceMut::new(local_buffer)],
             &[RemoteIoVec {
                 len,
-                base: usize::try_from(remote_addr).or(Err(Errno::EFAULT))?,
+                base: remote_addr,
             }],
         )
     }
 
+    /// Fallback method to allocate and read a `Zeroizing` buffer from `/proc/$pid/mem` when `process_vm_readv()` is unavailable.
+    ///
+    /// # Safety
+    ///
+    /// This function is unsafe because the request is not validated.
+    pub(crate) unsafe fn read_vec_zeroed_proc(
+        &self,
+        remote_addr: u64,
+        len: usize,
+    ) -> Result<Zeroizing<Vec<u8>>, Errno> {
+        let mut local_buffer = Zeroizing::new(Vec::new());
+
+        // Check for zero length and return an empty Vector.
+        if len == 0 {
+            return Ok(local_buffer);
+        }
+
+        // SAFETY: Check pointer against mmap_min_addr before allocation,
+        // but after length is zero check.
+        if remote_addr < *MMAP_MIN_ADDR {
+            return Err(Errno::EFAULT);
+        }
+
+        local_buffer.try_reserve(len).or(Err(Errno::ENOMEM))?;
+
+        // SAFETY: We are using `set_len(len)` after reserving enough
+        // capacity with `try_reserve(len)`. This is safe because the
+        // buffer was already allocated with enough memory to hold `len`
+        // elements, and we are not exceeding the reserved capacity. The
+        // memory is valid for `len` elements.
+        unsafe { local_buffer.set_len(len) };
+
+        let mut path = XPathBuf::from_pid(self.pid)?;
+        path.try_reserve(b"/mem".len()).or(Err(Errno::ENOMEM))?;
+        path.push(b"mem");
+
+        let mut file = safe_open_msym(PROC_FILE(), &path, OFlag::O_RDONLY, ResolveFlag::empty())
+            .map(File::from)
+            .or(Err(Errno::EACCES))?;
+        file.seek(SeekFrom::Start(remote_addr))
+            .or(Err(Errno::EACCES))?;
+
+        let mut nread = 0;
+        #[expect(clippy::arithmetic_side_effects)]
+        while nread < len {
+            match file.read(&mut local_buffer[nread..]) {
+                Ok(0) => return Err(Errno::EACCES),
+                Ok(n) => nread += n,
+                Err(ref e) if e.kind() == io::ErrorKind::Interrupted => {}
+                Err(_) => return Err(Errno::EACCES),
+            }
+        }
+
+        // SAFETY: nread is returned by the Linux kernel.
+        unsafe { local_buffer.set_len(nread) };
+        local_buffer.shrink_to_fit();
+
+        Ok(local_buffer)
+    }
+
+    /// Fallback method to allocate and read a buffer from `/proc/$pid/mem` when `process_vm_readv()` is unavailable.
+    ///
+    /// # Safety
+    ///
+    /// This function is unsafe because the request is not validated.
+    pub(crate) unsafe fn read_vec_proc(
+        &self,
+        remote_addr: u64,
+        len: usize,
+    ) -> Result<Vec<u8>, Errno> {
+        let mut local_buffer = Vec::new();
+
+        // Check for zero length and return an empty Vector.
+        if len == 0 {
+            return Ok(local_buffer);
+        }
+
+        // SAFETY: Check pointer against mmap_min_addr before allocation,
+        // but after length is zero check.
+        if remote_addr < *MMAP_MIN_ADDR {
+            return Err(Errno::EFAULT);
+        }
+
+        local_buffer.try_reserve(len).or(Err(Errno::ENOMEM))?;
+
+        // SAFETY: We are using `set_len(len)` after reserving enough
+        // capacity with `try_reserve(len)`. This is safe because the
+        // buffer was already allocated with enough memory to hold `len`
+        // elements, and we are not exceeding the reserved capacity. The
+        // memory is valid for `len` elements.
+        unsafe { local_buffer.set_len(len) };
+
+        let mut path = XPathBuf::from_pid(self.pid)?;
+        path.try_reserve(b"/mem".len()).or(Err(Errno::ENOMEM))?;
+        path.push(b"mem");
+
+        let mut file = safe_open_msym(PROC_FILE(), &path, OFlag::O_RDONLY, ResolveFlag::empty())
+            .map(File::from)
+            .or(Err(Errno::EACCES))?;
+        file.seek(SeekFrom::Start(remote_addr))
+            .or(Err(Errno::EACCES))?;
+
+        let mut nread = 0;
+        #[expect(clippy::arithmetic_side_effects)]
+        while nread < len {
+            match file.read(&mut local_buffer[nread..]) {
+                Ok(0) => return Err(Errno::EACCES),
+                Ok(n) => nread += n,
+                Err(ref e) if e.kind() == io::ErrorKind::Interrupted => {}
+                Err(_) => return Err(Errno::EACCES),
+            }
+        }
+
+        // SAFETY: nread is returned by the Linux kernel.
+        unsafe { local_buffer.set_len(nread) };
+        local_buffer.shrink_to_fit();
+
+        Ok(local_buffer)
+    }
+
     /// Fallback method to read data from `/proc/$pid/mem` when `process_vm_readv()` is unavailable.
     ///
     /// # Safety
@@ -457,13 +682,15 @@ impl RemoteProcess {
         &self,
         local_buffer: &mut [u8],
         remote_addr: u64,
+        len: usize,
     ) -> Result<usize, Errno> {
         // SAFETY: Check pointer against mmap_min_addr.
         if remote_addr < *MMAP_MIN_ADDR {
             return Err(Errno::EFAULT);
         }
 
-        let mut path = XPathBuf::from_pid(self.pid);
+        let mut path = XPathBuf::from_pid(self.pid)?;
+        path.try_reserve(b"/mem".len()).or(Err(Errno::ENOMEM))?;
         path.push(b"mem");
 
         let mut file = safe_open_msym(PROC_FILE(), &path, OFlag::O_RDONLY, ResolveFlag::empty())
@@ -474,7 +701,7 @@ impl RemoteProcess {
 
         let mut nread = 0;
         #[expect(clippy::arithmetic_side_effects)]
-        while nread < local_buffer.len() {
+        while nread < len {
             match file.read(&mut local_buffer[nread..]) {
                 Ok(0) => return Err(Errno::EACCES),
                 Ok(n) => nread += n,
@@ -496,13 +723,7 @@ impl RemoteProcess {
         local_buffer: &[u8],
         remote_addr: u64,
     ) -> Result<usize, Errno> {
-        static FORCE_PROC: LazyLock<bool> =
-            LazyLock::new(|| std::env::var_os(ENV_NO_CROSS_MEMORY_ATTACH).is_some());
-        // SAFETY: Falling back to proc_pid_mem(5) is insecure,
-        // and is no longer the default as of 3.32.6.
-        static PROC_MEM_F: LazyLock<bool> =
-            LazyLock::new(|| std::env::var_os(ENV_PROC_PID_MEM_FALLBACK).is_some());
-        if *FORCE_PROC || (*PROC_MEM_F && !*HAVE_CROSS_MEMORY_ATTACH) {
+        if Sandbox::use_proc_pid_mem() {
             return self.write_mem_proc(local_buffer, remote_addr);
         }
 
@@ -542,7 +763,8 @@ impl RemoteProcess {
             return Ok(0);
         }
 
-        let mut path = XPathBuf::from_pid(self.pid);
+        let mut path = XPathBuf::from_pid(self.pid)?;
+        path.try_reserve(b"/mem".len()).or(Err(Errno::ENOMEM))?;
         path.push(b"mem");
 
         let mut file = safe_open_msym(PROC_FILE(), &path, OFlag::O_WRONLY, ResolveFlag::empty())
@@ -593,7 +815,7 @@ impl RemoteProcess {
             // SAFETY:
             // 1. Assume error on zero-read.
             // 2. Validate the request after memory read.
-            let len = unsafe { self.read_mem(ptr, addr + off as u64) }?;
+            let len = unsafe { self.read_mem(ptr, addr + off as u64, len) }?;
             if len == 0 {
                 return Err(Errno::EFAULT);
             }
@@ -608,7 +830,13 @@ impl RemoteProcess {
 
                 // Adjust to actual size up to NUL-byte.
                 off += nul;
-                return Ok(buf[..off].into());
+
+                // Allocate vector on heap gracefully.
+                let mut vec = Vec::new();
+                vec.try_reserve(off).or(Err(Errno::ENOMEM))?;
+                vec.extend_from_slice(&buf[..off]);
+
+                return Ok(vec.into());
             }
 
             off += len;
@@ -685,7 +913,7 @@ impl Serialize for UNotifyEventRequest {
             // specified address.
             if let Some(addr) = auxv.get(&(libc::AT_RANDOM as u64)) {
                 let mut at_random = [0u8; 16];
-                if *addr >= *MMAP_MIN_ADDR && self.read_mem(&mut at_random, *addr).is_ok() {
+                if *addr >= *MMAP_MIN_ADDR && self.read_mem(&mut at_random, *addr, 16).is_ok() {
                     map.serialize_entry("at_random", &HEXLOWER.encode(&at_random))?;
                 }
             }
@@ -721,12 +949,12 @@ impl Serialize for UNotifyEventRequest {
         let mut ip_read = false;
         let mut sp_read = false;
 
-        if ip >= *MMAP_MIN_ADDR && self.read_mem(&mut ip_mem, ip).is_ok() {
+        if ip >= *MMAP_MIN_ADDR && self.read_mem(&mut ip_mem, ip, 64).is_ok() {
             ip_read = true;
         }
 
         if let Some(sp) = sp {
-            if sp >= *MMAP_MIN_ADDR && self.read_mem(&mut sp_mem, sp).is_ok() {
+            if sp >= *MMAP_MIN_ADDR && self.read_mem(&mut sp_mem, sp, 64).is_ok() {
                 sp_read = true;
             }
         }
@@ -931,16 +1159,13 @@ impl UNotifyEventRequest {
     ///
     /// Name must be a NUL-terminated string or `Err(Errno::ERANGE)` is returned.
     pub(crate) fn read_xattr(&self, addr: u64) -> Result<CString, Errno> {
-        let mut buf = Vec::new();
-
-        buf.try_reserve(XATTR_NAME_MAX).or(Err(Errno::ENOMEM))?;
-        buf.resize(XATTR_NAME_MAX, 0);
-
-        let mut n = self.read_mem(&mut buf, addr)?;
-
-        let z = memchr(0, &buf[..n]).ok_or(Errno::ERANGE)?;
-        n = n.min(z.checked_add(1).ok_or(Errno::ERANGE)?);
-        buf.truncate(n);
+        let mut buf = self.read_vec(addr, XATTR_NAME_MAX)?;
+        let z = memchr(0, &buf)
+            .ok_or(Errno::ERANGE)?
+            .checked_add(1)
+            .ok_or(Errno::ERANGE)?;
+        buf.truncate(z);
+        buf.shrink_to_fit();
 
         // Check for empty name.
         let len = buf.len(); // Includes NUL-byte.
@@ -1000,7 +1225,7 @@ impl UNotifyEventRequest {
             // Read remote memory.
             //
             // SAFETY: The request is going to be validated.
-            let n = unsafe { process.read_mem(slice, read_addr) }?;
+            let n = unsafe { process.read_mem(slice, read_addr, slice.len()) }?;
 
             // SAFETY: Assume error on zero-read.
             if n == 0 {
@@ -1033,25 +1258,35 @@ impl UNotifyEventRequest {
     /// Read the `libc::open_how` struct from process memory
     /// at the given address and size.
     pub(crate) fn remote_ohow(&self, addr: u64, size: u64) -> Result<libc::open_how, Errno> {
-        const OPEN_HOW_SIZE: usize = std::mem::size_of::<libc::open_how>();
-        const OPEN_HOW_ALIGN: usize = std::mem::align_of::<libc::open_how>();
+        const OPEN_HOW_SIZE_VER0: usize = 24;
+        const OPEN_HOW_SIZE_LATEST: usize = size_of::<libc::open_how>();
 
         // SAFETY: Validate size argument.
         let size = usize::try_from(size).or(Err(Errno::EINVAL))?;
-        match size.cmp(&OPEN_HOW_SIZE) {
-            Ordering::Equal => {}
-            Ordering::Less => return Err(Errno::EINVAL),
-            Ordering::Greater if size % OPEN_HOW_ALIGN == 0 => return Err(Errno::E2BIG),
-            Ordering::Greater => return Err(Errno::EFAULT),
-        };
+        if size < OPEN_HOW_SIZE_VER0 {
+            return Err(Errno::EINVAL);
+        }
+        if size as u64 > *PAGE_SIZE {
+            return Err(Errno::E2BIG);
+        }
 
         // SAFETY: Validate address argument.
         if addr < *MMAP_MIN_ADDR {
             return Err(Errno::EFAULT);
         }
 
-        let mut buf = [0u8; OPEN_HOW_SIZE];
-        self.read_mem(&mut buf, addr)?;
+        // Allocate buffer.
+        // Size is already capped to page size.
+        let raw = self.read_vec(addr, size)?;
+
+        // SAFETY: Verify trailing bytes are zero; otherwise E2BIG.
+        if raw.iter().skip(OPEN_HOW_SIZE_LATEST).any(|&b| b != 0) {
+            return Err(Errno::E2BIG);
+        }
+
+        let mut buf = [0u8; OPEN_HOW_SIZE_LATEST];
+        let len = buf.len().min(size);
+        buf[..len].copy_from_slice(&raw[..len]);
 
         // SAFETY: The following unsafe block assumes that:
         // 1. The memory layout of open_how in our Rust environment
@@ -1076,8 +1311,9 @@ impl UNotifyEventRequest {
             return Err(Errno::EFAULT);
         }
 
-        let mut buf = [0u8; std::mem::size_of::<libc::utimbuf>()];
-        self.read_mem(&mut buf, addr)?;
+        const LEN: usize = size_of::<libc::utimbuf>();
+        let mut buf = [0u8; LEN];
+        self.read_mem(&mut buf, addr, LEN)?;
 
         // SAFETY: The following unsafe block assumes that:
         // 1. The memory layout of utimbuf in our Rust environment
@@ -1107,8 +1343,9 @@ impl UNotifyEventRequest {
             return Err(Errno::EFAULT);
         }
 
-        let mut buf = [0u8; std::mem::size_of::<libc::timeval>() * 2];
-        self.read_mem(&mut buf, addr)?;
+        const LEN: usize = size_of::<libc::timeval>() * 2;
+        let mut buf = [0u8; LEN];
+        self.read_mem(&mut buf, addr, LEN)?;
 
         // SAFETY: The following unsafe block assumes that:
         // 1. The memory layout of timeval in our Rust environment
@@ -1150,8 +1387,9 @@ impl UNotifyEventRequest {
             return Err(Errno::EFAULT);
         }
 
-        let mut buf = [0u8; std::mem::size_of::<TimeSpec32>()];
-        self.read_mem(&mut buf, addr)?;
+        const LEN: usize = size_of::<TimeSpec32>();
+        let mut buf = [0u8; LEN];
+        self.read_mem(&mut buf, addr, LEN)?;
 
         // SAFETY: The following unsafe block assumes that:
         // 1. The memory layout of timespec in our Rust environment
@@ -1186,8 +1424,9 @@ impl UNotifyEventRequest {
             return Err(Errno::EFAULT);
         }
 
-        let mut buf = [0u8; std::mem::size_of::<TimeSpec64>()];
-        self.read_mem(&mut buf, addr)?;
+        const LEN: usize = size_of::<TimeSpec64>();
+        let mut buf = [0u8; LEN];
+        self.read_mem(&mut buf, addr, LEN)?;
 
         // SAFETY: The following unsafe block assumes that:
         // 1. The memory layout of timespec in our Rust environment
@@ -1220,8 +1459,9 @@ impl UNotifyEventRequest {
             return Err(Errno::EFAULT);
         }
 
-        let mut buf = [0u8; std::mem::size_of::<TimeSpec32>() * 2];
-        self.read_mem(&mut buf, addr)?;
+        const LEN: usize = size_of::<TimeSpec32>() * 2;
+        let mut buf = [0u8; LEN];
+        self.read_mem(&mut buf, addr, LEN)?;
 
         // SAFETY: The following unsafe block assumes that:
         // 1. The memory layout of timespec in our Rust environment
@@ -1262,8 +1502,9 @@ impl UNotifyEventRequest {
             return Err(Errno::EFAULT);
         }
 
-        let mut buf = [0u8; std::mem::size_of::<TimeSpec64>() * 2];
-        self.read_mem(&mut buf, addr)?;
+        const LEN: usize = size_of::<TimeSpec64>() * 2;
+        let mut buf = [0u8; LEN];
+        self.read_mem(&mut buf, addr, LEN)?;
 
         // SAFETY: The following unsafe block assumes that:
         // 1. The memory layout of timespec in our Rust environment
@@ -1300,13 +1541,12 @@ impl UNotifyEventRequest {
         &self,
         sandbox: &SandboxGuard,
         arg: SysArg,
-        magic: bool,
     ) -> Result<(CanonicalPath<'b>, bool, bool), Errno> {
         let process = RemoteProcess::new(self.scmpreq.pid());
 
         // SAFETY: The request is validated by read_path.
         let (path, magic, doterr, empty_path) =
-            process.read_path(sandbox, self.scmpreq.data.args, arg, magic, Some(self))?;
+            process.read_path(sandbox, self.scmpreq.data.args, arg, Some(self))?;
 
         // Determine FD-only system calls.
         // We return EACCES rather than ENOENT for these.
@@ -1341,16 +1581,55 @@ impl UNotifyEventRequest {
         }
     }
 
+    /// Read a `Zeroizing` vector from remote process's memory with `process_vm_readv()`.
+    pub(crate) fn read_vec_zeroed(
+        &self,
+        remote_addr: u64,
+        len: usize,
+    ) -> Result<Zeroizing<Vec<u8>>, Errno> {
+        let process = RemoteProcess::new(self.scmpreq.pid());
+
+        // SAFETY: The request is validated.
+        match unsafe { process.read_vec_zeroed(remote_addr, len) } {
+            Ok(vec) => {
+                if self.is_valid() {
+                    Ok(vec)
+                } else {
+                    Err(Errno::ESRCH)
+                }
+            }
+            Err(errno) => Err(errno),
+        }
+    }
+
+    /// Read a vector from remote process's memory with `process_vm_readv()`.
+    pub(crate) fn read_vec(&self, remote_addr: u64, len: usize) -> Result<Vec<u8>, Errno> {
+        let process = RemoteProcess::new(self.scmpreq.pid());
+
+        // SAFETY: The request is validated.
+        match unsafe { process.read_vec(remote_addr, len) } {
+            Ok(vec) => {
+                if self.is_valid() {
+                    Ok(vec)
+                } else {
+                    Err(Errno::ESRCH)
+                }
+            }
+            Err(errno) => Err(errno),
+        }
+    }
+
     /// Read data from remote process's memory with `process_vm_readv()`.
     pub(crate) fn read_mem(
         &self,
         local_buffer: &mut [u8],
         remote_addr: u64,
+        len: usize,
     ) -> Result<usize, Errno> {
         let process = RemoteProcess::new(self.scmpreq.pid());
 
         // SAFETY: The request is validated.
-        match unsafe { process.read_mem(local_buffer, remote_addr) } {
+        match unsafe { process.read_mem(local_buffer, remote_addr, len) } {
             Ok(n) => {
                 if self.is_valid() {
                     Ok(n)
