@@ -54,7 +54,7 @@ use std::{
     env,
     env::VarError,
     ffi::{OsStr, OsString},
-    fs::{File, OpenOptions},
+    fs::OpenOptions,
     io::{stdout, BufWriter, Write},
     net::{IpAddr, SocketAddrV4, SocketAddrV6},
     os::{
@@ -68,18 +68,19 @@ use std::{
 
 use data_encoding::HEXLOWER;
 use libseccomp::{scmp_cmp, ScmpAction, ScmpFilterContext, ScmpSyscall};
+use memchr::arch::all::is_equal;
 use nix::{
     errno::Errno,
     fcntl::OFlag,
     libc::setdomainname,
-    pty::{grantpt, posix_openpt, unlockpt},
+    pty::{grantpt, unlockpt},
     sched::{unshare, CloneFlags},
     sys::{
         socket::{bind, socket, AddressFamily, SockFlag, SockType, SockaddrIn, SockaddrIn6},
         wait::{Id, WaitPidFlag},
     },
     time::{clock_gettime, ClockId},
-    unistd::{fork, getgid, getpid, getuid, isatty, sethostname, ForkResult},
+    unistd::{fork, getgid, getpid, getuid, isatty, sethostname, ForkResult, Pid},
 };
 use sendfd::SendWithFd;
 use syd::{
@@ -89,19 +90,22 @@ use syd::{
     confine::{confine_scmp_madvise, ExportMode},
     err::err2no,
     error,
-    fs::{closeexcept, format_clone_flags, format_clone_names, openpts, pidfd_open, set_cloexec},
-    hash::{get_at_random_hex, hash, HashAlgorithm, SydHashMap, SydHashSet},
+    fs::{
+        closeexcept, format_clone_flags, format_clone_names, openpt, openpts, pidfd_open,
+        set_cloexec,
+    },
+    hash::{get_at_random_hex, hash, HashAlgorithm, SydHashSet},
     hook::Supervisor,
     ignore_signals, info,
     landlock_policy::LandlockPolicy,
     log::log_init,
     loopback_set_up,
     path::XPathBuf,
+    proc::proc_write,
     sandbox::Sandbox,
     seal::ensure_sealed,
     set_sigpipe_dfl, syd_code_name, syd_info,
     syslog::LogLevel,
-    unshare::{GidMap, UidMap},
     warn, IgnoreSignalOpts,
 };
 
@@ -573,61 +577,68 @@ syd::main! {
         sandbox.set_unshare_pid(true);
     }
 
-    let pty_debug = env::var_os("SYD_PTY_DEBUG").is_some();
     let pty_child = if sandbox.has_pty()
         && isatty(std::io::stdin()).unwrap_or(false)
         && isatty(std::io::stdout()).unwrap_or(false)
     {
-        // Step 1: Create a PIDFd of this process and clear O_CLOEXEC.
+        // Create a PIDFd of this process and clear O_CLOEXEC.
         let pidfd = pidfd_open(getpid(), libc::PIDFD_NONBLOCK)?;
         set_cloexec(&pidfd, false)?;
 
-        // Step 2: Open a pseudoterminal device without O_CLOEXEC.
-        let pty_m = posix_openpt(OFlag::O_RDWR | OFlag::O_NOCTTY | OFlag::O_NOFOLLOW)?;
+        // Open a pseudoterminal device and clear O_CLOEXEC.
+        let pty_m = openpt(OFlag::O_RDWR | OFlag::O_NOCTTY)?;
+        set_cloexec(&pty_m, false)?;
 
-        // Grant access and unlock.
+        // Grant access to PTY and unlock.
         grantpt(&pty_m)?;
         unlockpt(&pty_m)?;
 
         // Open peer device.
+        // We are going to pass this end to the sandbox process.
+        // This uses TIOCGPTPEER ioctl(2) so O_NOFOLLOW is not needed.
         let pty_s = openpts(&pty_m, OFlag::O_RDWR | OFlag::O_NOCTTY | OFlag::O_CLOEXEC)?;
 
-        // Step 3: Prepare environment of the syd-pty process.
+        // Prepare environment of the syd-pty process.
         // Filter the environment variables to only include the list below:
         // 1. PATH
         // 2. LD_LIBRARY_PATH
-        // 3. SYD_PTY_DEBUG
-        // 4. SYD_PTY_RULES
-        let safe_env: SydHashMap<_, _> = env::vars_os()
-            .filter(|(key, _)| {
-                matches!(
-                    key.as_bytes(),
-                    b"PATH" | b"LD_LIBRARY_PATH" | b"SYD_PTY_DEBUG" | b"SYD_PTY_RULES"
-                )
-            })
-            .collect();
+        // 3. SYD_PTY_RULES
+        // We do not need to pass SYD_PTY_DEBUG because we use -d as needed.
+        let safe_env: &[&[u8]] = &[b"PATH", b"LD_LIBRARY_PATH", b"SYD_PTY_RULES"];
 
-        // Step 5: Spawn syd-pty process, and pass PTY main end to it.
-        let mut cmd = std::process::Command::new(
-            env::var_os("CARGO_BIN_EXE_syd-pty").unwrap_or(OsString::from("syd-pty")),
+        // Spawn syd-pty process, and pass PTY main end to it.
+        let pty_debug = env::var_os("SYD_PTY_DEBUG").is_some();
+        let mut cmd = if let Some(cmd) = env::var_os("CARGO_BIN_EXE_syd-pty") {
+            std::process::Command::new(cmd)
+        } else {
+            std::process::Command::new("syd-pty")
+        };
+        cmd.stdin(Stdio::inherit());
+        cmd.stdout(Stdio::inherit());
+        cmd.env_clear();
+        cmd.envs(env::vars_os()
+            .filter(|(key, _)| {
+                    safe_env.iter().any(|env| is_equal(key.as_bytes(), env))
+            })
         );
         if pty_debug {
             cmd.arg("-d");
+            cmd.stderr(Stdio::inherit());
+        } else {
+            cmd.stderr(Stdio::null());
         }
-        cmd.arg(format!("-p{}", pidfd.as_raw_fd()));
-        cmd.arg(format!("-i{}", pty_m.as_raw_fd()));
+        let mut buf = itoa::Buffer::new();
+        cmd.arg("-p");
+        cmd.arg(buf.format(pidfd.as_raw_fd()));
+        cmd.arg("-i");
+        cmd.arg(buf.format(pty_m.as_raw_fd()));
         if let Some(ws) = sandbox.pty_ws_x() {
-            cmd.arg(format!("-x{ws}"));
+            cmd.arg("-x");
+            cmd.arg(buf.format(ws));
         }
         if let Some(ws) = sandbox.pty_ws_y() {
-            cmd.arg(format!("-y{ws}"));
-        }
-        cmd.env_clear();
-        cmd.envs(safe_env);
-        cmd.stdin(Stdio::inherit());
-        cmd.stdout(Stdio::inherit());
-        if !pty_debug {
-            cmd.stderr(Stdio::null());
+            cmd.arg("-y");
+            cmd.arg(buf.format(ws));
         }
         if let Err(error) = cmd.spawn() {
             let errno = err2no(&error);
@@ -647,9 +658,8 @@ syd::main! {
                 "msg": "syd-pty is now forwarding terminal I/O");
         }
 
-        // Step 6: Pass the other end of the PTY pair
-        // to the sandbox process.
-        env::set_var(ENV_PTY_FD, pty_s.as_raw_fd().to_string());
+        // Pass the other end of the PTY pair to the sandbox process.
+        env::set_var(ENV_PTY_FD, buf.format(pty_s.as_raw_fd()));
         Some(pty_s)
     } else {
         env::remove_var(ENV_PTY_FD);
@@ -661,60 +671,60 @@ syd::main! {
         // sandbox/proxy:on implies unshare/net:1
         sandbox.set_unshare_net(true);
 
-        // Step 1: Create a PIDFd of this process.
+        // Create a PIDFd of this process.
         let pidfd = pidfd_open(getpid(), libc::PIDFD_NONBLOCK)?;
 
-        // Step 2: Create a UNIX socket pair.
+        // Create a UNIX socket pair.
         let (stream_parent, stream_child) = UnixStream::pair()?;
 
-        // Step 3: Unset the CLOEXEC flags on the file descriptors.
+        // Unset the CLOEXEC flags on the file descriptors.
         // PIDFds and Rust sockets are automatically CLOEXEC.
         set_cloexec(&pidfd, false)?;
         set_cloexec(&stream_parent, false)?;
 
-        // Step 4: Prepare environment of the syd-tor process.
+        // Prepare environment of the syd-tor process.
         // Filter the environment variables to only include the list below:
         // 1. PATH
         // 2. LD_LIBRARY_PATH
-        // 3. SYD_TOR_DEBUG
-        // 4. SYD_TOR_RULES
-        let safe_env: SydHashMap<_, _> = env::vars_os()
-            .filter(|(key, _)| {
-                matches!(
-                    key.as_bytes(),
-                    b"PATH" | b"LD_LIBRARY_PATH" | b"SYD_TOR_DEBUG" | b"SYD_TOR_RULES"
-                )
-            })
-            .collect();
+        // 3. SYD_TOR_RULES
+        // We do not need to pass SYD_TOR_DEBUG because we use -d as needed.
+        let safe_env: &[&[u8]] = &[b"PATH", b"LD_LIBRARY_PATH", b"SYD_TOR_RULES"];
 
-        // Step 5: Spawn syd-tor process outside the namespace.
+        // Spawn syd-tor process outside the namespace.
         // Pass one end of the socket-pair to it.
-        let mut cmd = std::process::Command::new(
-            env::var_os("CARGO_BIN_EXE_syd-tor").unwrap_or(OsString::from("syd-tor")),
+        let mut cmd = if let Some(cmd) = env::var_os("CARGO_BIN_EXE_syd-tor") {
+            std::process::Command::new(cmd)
+        } else {
+            std::process::Command::new("syd-tor")
+        };
+        cmd.process_group(0);
+        cmd.stdin(Stdio::inherit());
+        cmd.stdout(Stdio::inherit());
+        cmd.env_clear();
+        cmd.envs(env::vars_os()
+            .filter(|(key, _)| {
+                    safe_env.iter().any(|env| is_equal(key.as_bytes(), env))
+            })
         );
         if proxy_debug {
             cmd.arg("-d");
+            cmd.stderr(Stdio::inherit());
+        } else {
+            cmd.stderr(Stdio::null());
         }
-        cmd.arg(format!("-p{}", pidfd.as_raw_fd()));
-        cmd.arg(format!("-i{}", stream_parent.as_raw_fd()));
+        let mut buf = itoa::Buffer::new();
+        cmd.arg("-p");
+        cmd.arg(buf.format(pidfd.as_raw_fd()));
+        cmd.arg("-i");
+        cmd.arg(buf.format(stream_parent.as_raw_fd()));
         // proxy/ext/unix has precedence over proxy/ext/host.
         if let Some(ref proxy_ext_unix) = sandbox.proxy_ext_unix {
             cmd.arg("-u");
             cmd.arg(proxy_ext_unix);
         } else {
+            let addr = format!("{}:{}", sandbox.proxy_ext_addr, sandbox.proxy_ext_port);
             cmd.arg("-o");
-            cmd.arg(format!(
-                "{}:{}",
-                sandbox.proxy_ext_addr, sandbox.proxy_ext_port
-            ));
-        }
-        cmd.env_clear();
-        cmd.envs(safe_env);
-        cmd.process_group(0);
-        cmd.stdin(Stdio::null());
-        cmd.stdout(Stdio::null());
-        if !proxy_debug {
-            cmd.stderr(Stdio::null());
+            cmd.arg(addr);
         }
         if let Err(error) = cmd.spawn() {
             let errno = err2no(&error);
@@ -725,22 +735,15 @@ syd::main! {
             return Err(error.into());
         }
         drop(pidfd);
-
-        let proxy_repr = sandbox
-            .proxy_ext_unix
-            .as_ref()
-            .map(|proxy_ext_unix| proxy_ext_unix.to_string())
-            .unwrap_or_else(|| format!("{}!{}", sandbox.proxy_ext_addr, sandbox.proxy_ext_port));
         if proxy_debug {
             warn!("ctx": "spawn_proxy",
-                "msg": format!("proxy is now forwarding external traffic to {proxy_repr}"));
+                "msg": format!("proxy is now forwarding external traffic to {}", sandbox.proxy_repr()));
         } else {
             info!("ctx": "spawn_proxy",
-                "msg": format!("proxy is now forwarding external traffic to {proxy_repr}"));
+                "msg": format!("proxy is now forwarding external traffic to {}", sandbox.proxy_repr()));
         }
 
-        // Step 4: Pass the other end of the socket-pair
-        // to the new namespace.
+        // Pass the other end of the socket-pair to the new namespace.
         Some((stream_child, sandbox.proxy_port))
     } else {
         None
@@ -777,42 +780,14 @@ syd::main! {
         let gid = getgid().as_raw();
 
         let map_root = sandbox.flags.map_root();
-
         let uid_buf = {
-            let uid_maps = vec![
-                UidMap {
-                    inside_uid: if map_root { 0 } else { uid },
-                    outside_uid: uid,
-                    count: 1,
-                }, // Map the current user.
-            ];
             let mut buf = Vec::new();
-            for map in uid_maps {
-                writeln!(
-                    &mut buf,
-                    "{} {} {}",
-                    map.inside_uid, map.outside_uid, map.count
-                )?;
-            }
+            write!(&mut buf, "{} {uid} 1", if map_root { 0 } else { uid })?;
             buf
         };
-
         let gid_buf = {
-            let gid_maps = vec![
-                GidMap {
-                    inside_gid: if map_root { 0 } else { gid },
-                    outside_gid: gid,
-                    count: 1,
-                }, // Map the current group.
-            ];
             let mut buf = Vec::new();
-            for map in gid_maps {
-                writeln!(
-                    &mut buf,
-                    "{} {} {}",
-                    map.inside_gid, map.outside_gid, map.count
-                )?;
-            }
+            write!(&mut buf, "{} {gid} 1", if map_root { 0 } else { gid })?;
             buf
         };
         Some((uid_buf, gid_buf))
@@ -846,12 +821,12 @@ syd::main! {
     }
 
     // Write uid/gid map for user namespace.
-    #[expect(clippy::disallowed_methods)]
-    if let Some((ref uid_buf, ref gid_buf)) = id_buf {
+    let pid = Pid::this();
+    if let Some((uid_buf, gid_buf)) = id_buf {
         // Write "deny" to /proc/self/setgroups before writing to gid_map.
-        File::create("/proc/self/setgroups").and_then(|mut f| f.write_all(b"deny"))?;
-        File::create("/proc/self/gid_map").and_then(|mut f| f.write_all(&gid_buf[..]))?;
-        File::create("/proc/self/uid_map").and_then(|mut f| f.write_all(&uid_buf[..]))?;
+        proc_write(pid, b"setgroups", b"deny")?;
+        proc_write(pid, b"gid_map", &gid_buf)?;
+        proc_write(pid, b"uid_map", &uid_buf)?;
 
         // Set inheritable mask and ambient caps to retain caps after execve(2).
         caps::securebits::set_keepcaps(true)?;
@@ -976,20 +951,35 @@ syd::main! {
     }
 
     if sandbox.flags.unshare_time() {
-        #[expect(clippy::disallowed_methods)]
-        if let Some(offset) = sandbox.time {
-            let mut file = File::create("/proc/self/timens_offsets")?;
-            let contents = format!("monotonic {offset} 0\nboottime {offset} 0\n");
-            file.write_all(contents.as_bytes())?;
-            info!("ctx": "set_boot_time", "msg": "set boot time in time namespace");
-        } else if let Ok(boottime) = clock_gettime(ClockId::CLOCK_BOOTTIME) {
-            // Set uptime to 0 for time namespace. Ignore errors.
-            if let Ok(mut file) = File::create("/proc/self/timens_offsets") {
-                #[expect(clippy::arithmetic_side_effects)]
-                let contents = format!("boottime {} 0\n", -boottime.tv_sec());
-                let _ = file.write_all(contents.as_bytes());
-            }
-            info!("ctx": "set_boot_time", "msg": "reset boot time in time namespace");
+        let boffset = if let Some(boffset) = sandbox.boottime {
+            boffset
+        } else {
+            let btime = clock_gettime(ClockId::CLOCK_BOOTTIME)?;
+            // Into is necessary on 32-bit.
+            #[expect(clippy::useless_conversion)]
+            btime.tv_sec().checked_neg().ok_or(Errno::EOVERFLOW)?.into()
+        };
+        let moffset = if let Some(moffset) = sandbox.monotime {
+            moffset
+        } else {
+            let mtime = clock_gettime(ClockId::CLOCK_MONOTONIC)?;
+            // Into is necessary on 32-bit.
+            #[expect(clippy::useless_conversion)]
+            mtime.tv_sec().checked_neg().ok_or(Errno::EOVERFLOW)?.into()
+        };
+
+        if boffset != 0 || moffset != 0 {
+            // <clock-id> <offset-secs> <offset-nanosecs>
+            let mut buf = itoa::Buffer::new();
+            let mut data = "monotonic ".to_string();
+            data.push_str(buf.format(moffset));
+            data.push_str(" 0\nboottime ");
+            data.push_str(buf.format(boffset));
+            data.push_str(" 0\n");
+
+            proc_write(pid, b"timens_offsets", data.as_bytes())?;
+            info!("ctx": "set_boot_time", "msg": "set boot time in time namespace",
+                "timens_offsets": [moffset, boffset]);
         }
     }
 
@@ -1103,7 +1093,7 @@ syd::main! {
         #[cfg(feature = "prof")]
         "gettid",
     ];
-    for name in allow_call {
+    for name in allow_call.iter().chain(VDSO_SYSCALLS) {
         if let Ok(syscall) = ScmpSyscall::from_name(name) {
             ctx.add_rule(ScmpAction::Allow, syscall)?;
         }
