@@ -46,7 +46,7 @@ use crate::{
     cookie::safe_memfd_create,
     debug,
     err::err2no,
-    fs::{pidfd_send_signal, safe_clone, seal_memfd, set_cloexec, set_nonblock},
+    fs::{pidfd_send_signal, safe_clone, seal_memfd_all, set_cloexec, set_nonblock},
     get_user_home, get_user_name,
     hash::SydHashSet,
     landlock::RulesetStatus,
@@ -335,7 +335,7 @@ impl WordExp<'_> {
         file.write_all(b"\nshift\nprintf '%s ' \"$@\"\n")?;
 
         // Seal memfd for future writes.
-        seal_memfd(&file)?;
+        seal_memfd_all(&file)?;
         debug!("ctx": "expand",
             "msg": format!("sealed memory-file {} against grows, shrinks and writes",
                 file.as_raw_fd()));
