@@ -47,12 +47,13 @@ use crate::{
     debug,
     err::{err2no, SydJoinHandle, SydResult},
     error,
-    fs::{seal_memfd, FileInfo},
+    fs::seal_memfd_all,
     hash::{
         aes_ctr_enc, aes_ctr_init, hmac_sha256_feed, hmac_sha256_fini, hmac_sha256_init,
         SydHashMap, BLOCK_SIZE, HMAC_TAG_SIZE, IV, IV_SIZE, SYD3_HDR_SIZE,
     },
     info,
+    lookup::FileInfo,
     ofd::lock_fd,
     path::{XPath, XPathBuf},
     retry::{retry_on_eintr, retry_on_intr},
@@ -193,6 +194,9 @@ impl AesWorker {
         // We need this because std::thread::spawn has unwanted
         // side-effects such as opening /sys/devices/system/cpu/online
         // on some architectures.
+        //
+        // Note, we avoid this when profiling is enabled,
+        // as gperf requires it to write profiling data.
         for sysname in [
             "open",
             "openat",
@@ -204,7 +208,12 @@ impl AesWorker {
         ] {
             match ScmpSyscall::from_name(sysname) {
                 Ok(syscall) => {
-                    ctx.add_rule(ScmpAction::Errno(Errno::ENOSYS as i32), syscall)?;
+                    let action = if !cfg!(feature = "prof") {
+                        ScmpAction::Errno(Errno::ENOSYS as i32)
+                    } else {
+                        ScmpAction::Allow
+                    };
+                    ctx.add_rule(action, syscall)?;
                 }
                 Err(_) => {
                     info!("ctx": "confine", "op": "allow_aes_syscall",
@@ -626,7 +635,7 @@ impl AesWorker {
     ) -> SydResult<()> {
         // Seal memfd to ensure no further writes happen.
         if memfd {
-            seal_memfd(&crypt_fd)?;
+            seal_memfd_all(&crypt_fd)?;
             debug!("ctx": "aes", "op": "seal_memfd",
                 "path": crypt_path, "memfd": memfd,
                 "msg": "sealed plaintext memfd to prevent seals, writes, shrinks and grows");
